/*
	Copyright 2007 Oscar Rodriguez Parra

	This file is part of fonerawd.
	
	fonerawd is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	Foobar is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with fonerawd; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <termios.h>
#include <stdio.h>
#include <string.h>
#include "fonerawd.h"

int main( void )
{
	int puerto;
	char gps[ 25 ] = "";
	pid_t pid;
	pid_t sid;

	/* Activamos el manejo de mensajes al demonio de syslogd */
	openlog( "fonerawd", LOG_CONS | LOG_PID, LOG_DAEMON );
	/* Creamos un hilo de ejecucion */
	pid = fork();
	
	if( pid < 0 )
	{
		syslog( LOG_ERR, "No se pudo crear el proceso hijo" );
		exit( EXIT_FAILURE );
	}
	
	if( pid > 0 )
	{
		exit( EXIT_SUCCESS );
	}
	/* Convertimos el proceso hijo a demonio */
	sid = setsid();

	if( sid < 0 )
	{
		syslog( LOG_ERR, "No se pudo crear el demonio" );
		exit( EXIT_FAILURE );
	}
	Configuracion( &puerto );

	/* Bucle principal encargado de buscar la posicion GPS y las redes wifi */
	while( TRUE )
	{
		ProcesaGPS( &puerto, gps );
		ProcesaRedes( gps );
	}

	/* Aqui se supone que nunca llegaria */
	close( puerto );
	closelog();
	return EXIT_SUCCESS;
}

void Configuracion( int *puerto )
{
	/* Al ser un demonio no se dispone de la entrada y salida estandar
	 * y de salida de errores, y por ello se cierran */
	close( STDIN_FILENO );
	close( STDOUT_FILENO );
	close( STDERR_FILENO );
	/* Ponemos un umask definido por nosotros y no el heredado */
	umask( 0 );

        /* Creamos el directorio */
	if( ( mkdir( DIRECTORIO, 0755 ) == -1 ) && ( errno != EEXIST ) )
	{
		syslog( LOG_ERR, "No se pudo crear el directorio" );
		exit( EXIT_FAILURE );
	}
	/* Entramos en el directorio */
	chdir( DIRECTORIO );
	ConfiguraPuerto( puerto );
}

void ConfiguraPuerto( int *puerto )
{
	struct termios newtio;
	
	/* Abrimos el puerto serie */
	*puerto = open( MODEMDEVICE, O_RDWR | O_NOCTTY );
	
	if( *puerto == -1 ) 
	{
		syslog( LOG_ERR, "No se pudo abrir el puerto serie" );
		exit( EXIT_FAILURE);
	}
	/* Establecemos los nuevos valores del puerto serie */
	newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNPAR | ICRNL;
	newtio.c_oflag = 0;
	newtio.c_lflag = ICANON;
	newtio.c_cc[VINTR]    = 0;     /* Ctrl-c */ 
	newtio.c_cc[VQUIT]    = 0;     /* Ctrl-\ */
	newtio.c_cc[VERASE]   = 0;     /* del */
	newtio.c_cc[VKILL]    = 0;     /* @ */
	newtio.c_cc[VEOF]     = 4;     /* Ctrl-d */
	newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused */
	newtio.c_cc[VMIN]     = 1;     /* blocking read until 1 character arrives */
	newtio.c_cc[VSWTC]    = 0;     /* '\0' */
	newtio.c_cc[VSTART]   = 0;     /* Ctrl-q */ 
	newtio.c_cc[VSTOP]    = 0;     /* Ctrl-s */
	newtio.c_cc[VSUSP]    = 0;     /* Ctrl-z */
	newtio.c_cc[VEOL]     = 0;     /* '\0' */
	newtio.c_cc[VREPRINT] = 0;     /* Ctrl-r */
	newtio.c_cc[VDISCARD] = 0;     /* Ctrl-u */
	newtio.c_cc[VWERASE]  = 0;     /* Ctrl-w */
	newtio.c_cc[VLNEXT]   = 0;     /* Ctrl-v */
	newtio.c_cc[VEOL2]    = 0;     /* '\0' */
	tcflush( *puerto, TCIFLUSH );
	tcsetattr( *puerto, TCSANOW, &newtio );
}

void ProcesaGPS( int *puerto, char *gps )
{
	int res;
	char buffer[ 256 ];
	char *cadenaNMEA;
        fd_set lectura;
        struct timeval Expiracion;

        /* Inicializamos la lectura por multiplexacion */
	Expiracion.tv_usec = 0;
        Expiracion.tv_sec = 2;
	FD_ZERO( &lectura );
	FD_SET( *puerto, &lectura );

	/* Comprobamos si hay datos que leer */
	if( select( *puerto + 1, &lectura, NULL, NULL, &Expiracion ) > 0 )
	{
		/* Comprobamos si son los datos del puerto serie */
		if( FD_ISSET( *puerto, &lectura ) )
		{
			/* Leemos la cadena NMEA del puerto serie */
			res = read( *puerto, buffer, 255 );
			buffer[ res ] = '\0';
			/* Buscamos la sentencia RMC */
			cadenaNMEA = strstr( buffer, "$GPRMC" );
			
			if( cadenaNMEA != NULL )
			{
				GuardaGPS( cadenaNMEA, gps );
				/* Esto parece ser necesario para que tome los valores actualizados */
				tcflush( *puerto, TCIFLUSH );
			}
		}
	}
	else
	{
		strcpy( gps, "error" );
	}
}

void GuardaGPS( char *cadenaNMEA, char *gps )
{
	char *GPSactivo;
	int contador;
	char *posicion[ 2 ];
	char *punto[ 2 ];
	float grados;
	float minutos;
	float dato[ 2 ]; 

	/* Saltamos la cadena GPRMC */
	strtok( cadenaNMEA, "," );
	/* Saltamos la hora UTC */
	strtok( NULL, "," );
	GPSactivo = strtok( NULL, "," );

	/* Comprobamos que el GPS contiene una latitud y longitud validas */
	if( *GPSactivo == 'A' )
	{
		/* Recorremos los datos de latitud y longitud */
		for( contador = 0; contador < 2; contador++ )
		{
			/* Guardamos la posicion geografica */
			posicion[ contador ] = strtok( NULL, "," );
			/* Guardamos el punto cardinal */
			punto[ contador ] = strtok( NULL, "," );
			/* Extraemos los grados y los minutos */
			sscanf( posicion[ contador ], "%f", &dato[ contador ] );
			grados = floor( dato[ contador ] / 100 );
			minutos = dato[ contador ] - ( grados * 100 );
			/* Los convertimos a coordenadas */
			dato[ contador ] = grados + ( minutos / 60 );

			/* Si el punto es sur u oeste entonces el valor es negativo */
			if( *punto[ contador ] == 'S' || *punto[ contador ] == 'W' )
			{
				dato[ contador ] *= -1;
			}
		}
	}
	else
	{
		dato[ 0 ] = 0;
		dato[ 1 ] = 0;
	}
	/* Guardamos la cadena GPS formateada */
	sprintf( gps, "%3.5f %3.5f", dato[ 0 ], dato[ 1 ] );
}

void ProcesaRedes( char *gps )
{
	wireless_scan_head wsh;
	wireless_scan *ws;
	int skfd;
	int ret;

	/* Creamos un socket necesario para el escaneo */
	skfd = socket( AF_INET, SOCK_DGRAM, 0 );
	
	if( skfd == -1 )
	{
		syslog( LOG_ERR, "No se pudo crear el socket" );
		exit( EXIT_FAILURE );
	}
	/* Escaneamos para recuperar la lista de redes wifi */
	ret = iw_scan( skfd, DISPOSITIVO, WIRELESS_EXT, &wsh );

	/* Si no hay datos esperamos un segundo y repetimos */
	if( ( ret == -1 ) && ( errno == ENODATA ) )
       	{
		sleep( 1 );
		ret = iw_scan( skfd, DISPOSITIVO, WIRELESS_EXT, &wsh );
		
		if( ret == -1 )
		{
			syslog( LOG_ERR, "No se pudo recuperar la lista de redes" );
			exit( EXIT_FAILURE );
		}
	}
	ws = wsh.result;
	/* Guardamos la lista de redes wifi */
	GuardaRedes( ws, gps );
	close( skfd );
}

void GuardaRedes( wireless_scan *ws, char *gps )
{
	char direccion_mac[ 13 ];
	FILE *fichero;
	wireless_scan *anterior;

        /* Recorremos la lista de redes wifi y por cada una creamos un fichero
	 * con su direccion MAC */
	while( ws )
	{
		fichero = fopen( Devuelve_MAC( ws, direccion_mac), "w" );

                if( fichero == NULL )
		{
			syslog( LOG_ERR, "No se pudo crear el fichero" );
                        exit( EXIT_FAILURE );
                }
                /* En el fichero guardamos su ESSID, su tipo, si esta encriptada y 
		 * su posicion GPS */
                fprintf( fichero, "%s|%d|%d|%s\n",
                                ws->b.essid,
				ws->b.mode,
				( ( ws->b.key_flags & 0x8000 ) == 0 ),
                                gps );
                fflush( fichero );
                fclose( fichero );
		anterior = ws;
                ws = ws->next;
		free( anterior );
        }
}

char *Devuelve_MAC( wireless_scan *ws, char *direccion_mac )
{
	char tabla[]={'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
	int indice;
	char *puntero;

	puntero = direccion_mac;
	
	/* Recorremos los 6 bytes de la direccion MAC y los convertimos a una
	 * cadena hexadecimal */
	for( indice = 0; indice < 6; indice ++, puntero += 2 )
	{
		*( puntero ) = tabla[ ( ( ws->ap_addr.sa_data[ indice ] & 0xF0 ) >> 4 ) ] ;
		*( puntero + 1 ) = tabla[ ( ws->ap_addr.sa_data[ indice ] & 0x0F ) ] ;
	}
	*( puntero ) = '\0';
	
	return direccion_mac;
}
